package com.linbit.linstor.core.apicallhandler.controller;

import com.linbit.ChildProcessTimeoutException;
import com.linbit.ImplementationError;
import com.linbit.InvalidNameException;
import com.linbit.extproc.DaemonHandler;
import com.linbit.extproc.ExtCmd;
import com.linbit.extproc.ExtCmd.OutputData;
import com.linbit.extproc.ExtCmdFactory;
import com.linbit.extproc.ExtCmdFailedException;
import com.linbit.extproc.OutputProxy.EOFEvent;
import com.linbit.extproc.OutputProxy.Event;
import com.linbit.extproc.OutputProxy.ExceptionEvent;
import com.linbit.extproc.OutputProxy.StdErrEvent;
import com.linbit.extproc.OutputProxy.StdOutEvent;
import com.linbit.linstor.InternalApiConsts;
import com.linbit.linstor.SosReportType;
import com.linbit.linstor.annotation.PeerContext;
import com.linbit.linstor.api.ApiCallRc;
import com.linbit.linstor.api.ApiCallRcImpl;
import com.linbit.linstor.api.ApiConsts;
import com.linbit.linstor.api.interfaces.serializer.CtrlStltSerializer;
import com.linbit.linstor.api.pojo.RequestFilePojo;
import com.linbit.linstor.core.LinStor;
import com.linbit.linstor.core.apicallhandler.ScopeRunner;
import com.linbit.linstor.core.apicallhandler.controller.internal.CtrlSatelliteUpdateCaller;
import com.linbit.linstor.core.apicallhandler.response.ApiAccessDeniedException;
import com.linbit.linstor.core.apicallhandler.response.ApiRcException;
import com.linbit.linstor.core.cfg.CtrlConfig;
import com.linbit.linstor.core.cfg.LinstorConfig;
import com.linbit.linstor.core.identifier.NodeName;
import com.linbit.linstor.core.objects.Node;
import com.linbit.linstor.core.objects.ResourceDefinition;
import com.linbit.linstor.core.repository.NodeRepository;
import com.linbit.linstor.core.repository.ResourceDefinitionRepository;
import com.linbit.linstor.dbdrivers.DatabaseException;
import com.linbit.linstor.dbdrivers.DbEngine;
import com.linbit.linstor.logging.ErrorReporter;
import com.linbit.linstor.netcom.Peer;
import com.linbit.linstor.netcom.PeerNotConnectedException;
import com.linbit.linstor.proto.responses.FileOuterClass.File;
import com.linbit.linstor.proto.responses.MsgSosReportFilesReplyOuterClass.MsgSosReportFilesReply;
import com.linbit.linstor.proto.responses.MsgSosReportListReplyOuterClass.FileInfo;
import com.linbit.linstor.proto.responses.MsgSosReportListReplyOuterClass.MsgSosReportListReply;
import com.linbit.linstor.security.AccessContext;
import com.linbit.linstor.security.AccessDeniedException;
import com.linbit.linstor.storage.kinds.ExtTools;
import com.linbit.linstor.storage.kinds.ExtToolsInfo;
import com.linbit.linstor.utils.FileUtils;
import com.linbit.locks.LockGuardFactory;
import com.linbit.locks.LockGuardFactory.LockObj;
import com.linbit.locks.LockGuardFactory.LockType;
import com.linbit.utils.FileCollector;
import com.linbit.utils.StringUtils;

import javax.inject.Provider;
import javax.inject.Singleton;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.nio.file.attribute.FileTime;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TimeZone;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.google.inject.Inject;
import reactor.core.publisher.Flux;

@Singleton
public class CtrlSosReportApiCallHandler
{
    private static final String QUERY_FILE = "query-params";

    private static final int DFLT_DEQUE_CAPACITY = 10;

    private static final String SOS_PREFIX = "sos_";
    private static final String SOS_SUFFIX = ".tar.gz";

    /** 10MiB */
    private static final long MAX_PACKET_SIZE = 10 * (1 << 20);
    private final DateFormat sdf = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");
    private final DateFormat sdfUtc = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");
    private final Provider<AccessContext> peerAccCtx;
    private final ScopeRunner scopeRunner;
    private final LockGuardFactory lockGuardFactory;
    private final NodeRepository nodeRepository;
    private final ResourceDefinitionRepository rscDfnRepo;
    private final CtrlStltSerializer stltComSerializer;
    private final ErrorReporter errorReporter;
    private final ExtCmdFactory extCmdFactory;
    private final CtrlConfig ctrlCfg;
    private final DbEngine db;

    @Inject
    public CtrlSosReportApiCallHandler(
        ErrorReporter errorReporterRef,
        @PeerContext
        Provider<AccessContext> peerAccCtxRef,
        ScopeRunner scopeRunnerRef,
        LockGuardFactory lockGuardFactoryRef,
        NodeRepository nodeRepositoryRef,
        ResourceDefinitionRepository rscDfnRepoRef,
        CtrlStltSerializer clientComSerializerRef,
        ExtCmdFactory extCmdFactoryRef,
        CtrlConfig ctrlCfgRef,
        DbEngine dbRef
    )
    {
        errorReporter = errorReporterRef;
        peerAccCtx = peerAccCtxRef;
        scopeRunner = scopeRunnerRef;
        lockGuardFactory = lockGuardFactoryRef;
        nodeRepository = nodeRepositoryRef;
        rscDfnRepo = rscDfnRepoRef;
        stltComSerializer = clientComSerializerRef;
        extCmdFactory = extCmdFactoryRef;
        ctrlCfg = ctrlCfgRef;
        db = dbRef;
        sdfUtc.setTimeZone(TimeZone.getTimeZone("UTC"));
    }

    /**
     * Overall workflow is as follows:
     *
     * First, the controller sends out requests to all satellite for a file-list (name, size and length; no content).
     * We simply assume that this file-list is per satellite smaller than 16MB.
     *
     * Next, in order to not exceed that 16MB limit, the controller requests the first files per list that sum of sizes
     * do not exceed 10MB.
     * The rest of the 16MB is a (hopefully more than generous) buffer for the overhead from Linstor-protocol and
     * protobuf, and others.
     *
     * The responses to those requests are stored on the controller directly in the destination files.
     *
     * Once all satellites finished the responses, the controller tar.gz's the collected files and sends a
     * cleanup-message to the satellites
     * (which deletes their temporary sos-directories).
     *
     * The resulting String in the flux is the name of the SOS-report.
     */
    public Flux<String> getSosReport(
        Set<String> nodes,
        Set<String> rscs,
        Set<String> excludeNodes,
        Date since,
        boolean includeCtrl,
        String queryParams
    )
    {
        Flux<String> ret;
        try
        {
            final Path tmpDir = Files.createTempDirectory("sos");
            makeDir(tmpDir.resolve("tmp"));

            String sosReportName = SOS_PREFIX + sdfUtc.format(new Date());

            ret = scopeRunner.fluxInTransactionlessScope(
                "Send SOS report queries",
                lockGuardFactory.buildDeferred(LockType.READ, LockObj.NODES_MAP, LockObj.RSC_DFN_MAP),
                () -> sendRequests(nodes, rscs, excludeNodes, tmpDir, sosReportName, since)
            ).flatMap(
                sosFileList -> scopeRunner.fluxInTransactionalScope(
                    "Receiving SOS FileList",
                    lockGuardFactory.buildDeferred(LockType.READ, LockObj.NODES_MAP),
                    () -> handleSosFileList(tmpDir, sosReportName, sosFileList)
                )
            ).transform(
                // until now we were working with Flux<ByteArrayInputStream>s (proto messages from the satellites), but
                // now we need to prepare for the answer to the satellite, which expects a single string of the sos
                // report name
                ignore -> ignore.thenMany(Flux.<String>empty())
            ).concatWith(
                scopeRunner.fluxInTransactionalScope(
                    "Finishing SOS report",
                    lockGuardFactory.buildDeferred(LockType.READ, LockObj.NODES_MAP, LockObj.RSC_DFN_MAP),
                    () -> finishReport(
                        nodes, rscs, excludeNodes, tmpDir, sosReportName, since, includeCtrl, queryParams
                    )
                )
            );
        }
        catch (IOException exc)
        {
            errorReporter.reportError(exc);
            ret = Flux.error(exc);
        }
        return ret;
    }

    /**
     * Sends the initial request of file-lists to all given satellites.
     * Next flux: process answers
     */
    private Flux<ByteArrayInputStream> sendRequests(
        Set<String> nodes,
        Set<String> rscs,
        Set<String> excludeNodes,
        Path tmpDir,
        String sosReportName,
        Date since
    )
        throws AccessDeniedException
    {
        Stream<Node> nodeStream = getNodeStreamForSosReport(nodes, rscs, excludeNodes);

        List<Flux<ByteArrayInputStream>> namesAndRequests = nodeStream
            .map(node -> prepareSosRequestApi(node, tmpDir, sosReportName, since))
            .collect(Collectors.toList());

        return Flux.fromIterable(namesAndRequests).flatMap(Function.identity());
    }

    /**
     * If the peer cannot be found from the given node, return a <code>Flux.empty()</code>. Otherwise return the Flux
     * from <code>peer.apiCall(...)</code>
     */
    private Flux<ByteArrayInputStream> prepareSosRequestApi(
        Node node,
        Path tmpDir,
        String sosReportName,
        Date since
    )
    {
        Peer peer = getPeer(node);
        Flux<ByteArrayInputStream> fluxReturn = Flux.empty();
        if (peer != null)
        {
            // already create the extTools file here, because you need the peer
            createExtToolsInfo(tmpDir.resolve("tmp/" + node.getName().displayValue), peer);
            byte[] msg = stltComSerializer.headerlessBuilder().requestSosReport(sosReportName, since).build();
            fluxReturn = peer.apiCall(InternalApiConsts.API_REQ_SOS_REPORT_FILE_LIST, msg)
                .onErrorResume(PeerNotConnectedException.class, ignored -> Flux.empty());
        }
        return fluxReturn;
    }

    private Peer getPeer(Node node)
    {
        Peer peer;
        try
        {
            peer = node.getPeer(peerAccCtx.get());
        }
        catch (AccessDeniedException exc)
        {
            throw new ApiAccessDeniedException(
                exc,
                "access peer for node '" + node.getName().displayValue + "'",
                ApiConsts.FAIL_ACC_DENIED_NODE
            );
        }
        return peer;
    }

    /**
     * Called when the satellite responses their file list (without content).
     * The ProtoFiles are converted into Pojos, empty files (0 bytes) are already created,
     * and <code>requestNextBatch</code> is called.
     */
    private Flux<ByteArrayInputStream> handleSosFileList(
        Path tmpDirRef,
        String sosReportName,
        ByteArrayInputStream sosFileListInputStreamRef
    )
    {
        MsgSosReportListReply msgSosReportListReply;
        Peer peer;
        try
        {
            msgSosReportListReply = MsgSosReportListReply.parseDelimitedFrom(sosFileListInputStreamRef);
            peer = getPeer(nodeRepository.get(peerAccCtx.get(), new NodeName(msgSosReportListReply.getNodeName())));
        }
        catch (IOException exc)
        {
            throw new ApiRcException(
                ApiCallRcImpl.simpleEntry(
                    ApiConsts.FAIL_UNKNOWN_ERROR,
                    "An IOException occurred while parsing SosFileList"
                ),
                exc
            );
        }
        catch (AccessDeniedException accDeniedExc)
        {
            throw new ApiAccessDeniedException(accDeniedExc, "accessing node", ApiConsts.FAIL_ACC_DENIED_NODE);
        }
        catch (InvalidNameException exc)
        {
            throw new ImplementationError(exc);
        }

        Flux<ByteArrayInputStream> flux;
        if (peer == null)
        {
            flux = Flux.empty();
        }
        else
        {
            Path sosDir = tmpDirRef.resolve("tmp/" + msgSosReportListReply.getNodeName());
            String sosReportDirOnStltStr = LinStor.SOS_REPORTS_DIR.resolve(sosReportName).toString();
            String sosDirStr = sosDir.toString();
            LinkedList<RequestFilePojo> filesToRequest = new LinkedList<>();
            for (FileInfo fileInfo : msgSosReportListReply.getFilesList())
            {
                if (fileInfo.getSize() == 0)
                {
                    // do not request files with 0 bytes. instead, just create the corresponding file right now

                    String fileName = fileInfo.getName();
                    if (fileName.startsWith(sosReportDirOnStltStr))
                    {
                        // move all '$stlt/var/lib/linstor.d/sos-report/$currentSosName/*' files to '$stlt/*'.
                        fileName = fileName.substring(sosReportDirOnStltStr.length());
                    }
                    errorReporter.logTrace(
                        "Not requesting %s for file %s as its size is 0 bytes. File %s created.",
                        peer.getNode().getName().displayValue,
                        fileInfo.getName(),
                        fileName
                    );

                    append(
                        concatPaths(sosDirStr, fileName),
                        new byte[0],
                        fileInfo.getTime()
                    );
                }
                else
                {
                    filesToRequest.add(new RequestFilePojo(fileInfo.getName(), 0, fileInfo.getSize()));
                }
            }
            if (!msgSosReportListReply.getErrorMessage().isEmpty())
            {
                append(
                    sosDir.resolve("sos.err"),
                    (msgSosReportListReply.getErrorMessage() + "\n").getBytes(),
                    System.currentTimeMillis()
                );
            }
            flux = requestNextBatch(tmpDirRef, sosReportName, peer, filesToRequest);
        }

        return flux;
    }

    /**
     * This method takes the original fileList and processes all elements. A fully processed entry is deleted from the
     * list. An element is fully processed if all of its bytes can be requested from the satellite. A file needs to be
     * split into multiple requests if the current request's size exceeds 10MiB. In that case, the partially requested
     * file is re-added to the list (at beginning, at index 0) with modified offset for the next processing.
     * The prepared batch is requested from the satellite and processed in the method <code>handleReceivedFiles</code>.
     * If the fileList is empty (i.e. the last batch was requested and processed), this method returns an empty Flux
     * finishing the Flux-loop.
     */
    private Flux<ByteArrayInputStream> requestNextBatch(
        Path tmpDirRef,
        String sosReportNameRef,
        Peer stltPeerRef,
        LinkedList<RequestFilePojo> filesToRequestRef
    )
    {
        long sumFileSizes = 0;
        ArrayList<RequestFilePojo> nextBatchToRequest = new ArrayList<>();
        while (!filesToRequestRef.isEmpty())
        {
            RequestFilePojo next = filesToRequestRef.removeFirst();
            long nextSize = next.length - next.offset;
            if (nextSize + sumFileSizes <= MAX_PACKET_SIZE)
            {
                // the request fits, but it might be the last part of an split request.
                RequestFilePojo lastPartialRequestFile = new RequestFilePojo(next.name, next.offset, nextSize);
                nextBatchToRequest.add(lastPartialRequestFile);
                sumFileSizes += nextSize;
                if (lastPartialRequestFile.offset == 0)
                {
                    errorReporter.logTrace(
                        "Requesting %s for file %s (total size: %d)",
                        stltPeerRef.getNode().getName().displayValue,
                        lastPartialRequestFile.name,
                        lastPartialRequestFile.length
                    );
                }
                else
                {
                    errorReporter.logTrace(
                        "Requesting %s for last part of file %s (offset: %d, length: %d, total size: %d)",
                        stltPeerRef.getNode().getName().displayValue,
                        lastPartialRequestFile.name,
                        lastPartialRequestFile.offset,
                        lastPartialRequestFile.length,
                        next.length
                    );
                }
            }
            else
            {
                long partSize = MAX_PACKET_SIZE - sumFileSizes;

                RequestFilePojo partialRequestFile = new RequestFilePojo(next.name, next.offset, partSize);
                nextBatchToRequest.add(partialRequestFile);
                errorReporter.logTrace(
                    "Requesting %s for file %s (offset: %d, length: %d, total size: %d)",
                    stltPeerRef.getNode().getName().displayValue,
                    partialRequestFile.name,
                    partialRequestFile.offset,
                    partialRequestFile.length,
                    next.length
                );

                // safe the partially requested file for next iteration
                RequestFilePojo nextContinueFrom = new RequestFilePojo(next.name, next.offset + partSize, next.length);
                filesToRequestRef.addFirst(nextContinueFrom);
                break;
            }
        }

        Flux<ByteArrayInputStream> ret;
        if (!nextBatchToRequest.isEmpty())
        {
            ret = stltPeerRef.apiCall(
                InternalApiConsts.API_REQ_SOS_REPORT_FILES,
                stltComSerializer.headerlessBuilder()
                    .requestSosReportFiles(
                        sosReportNameRef,
                        nextBatchToRequest
                    )
                    .build()
            ).flatMap(
                response -> handleReceivedFiles(tmpDirRef, sosReportNameRef, stltPeerRef, filesToRequestRef, response)
            )
                .onErrorResume(PeerNotConnectedException.class, ignored -> Flux.empty());
        }
        else
        {
            ret = Flux.empty(); // we are done with requests to this satellite
        }
        return ret;
    }

    /**
     * Processes the requested batch (writes the content to the corresponding files) and calls the previous method
     * (<code>requestNextBatch</code>) for the next batch.
     *
     * If a processed file starts with /var/lib/linstor.d/sos-report/$currentSosName/, instead of creating a
     * satellite-local directory with that name, we simplycreate the file into the satellite-root directory. In other
     * words all '$stlt/var/lib/linstor.d/sos-report/$currentSosName/*' files are going to land in '$stlt/*'.
     */
    private Flux<ByteArrayInputStream> handleReceivedFiles(
        Path tmpDirRef,
        String sosReportNameRef,
        Peer stltPeerRef,
        LinkedList<RequestFilePojo> filesToRequestRef,
        ByteArrayInputStream responseRef
    )
    {
        MsgSosReportFilesReply protoFilesReply;
        try
        {
            protoFilesReply = MsgSosReportFilesReply.parseDelimitedFrom(responseRef);
        }
        catch (IOException exc)
        {
            throw new ApiRcException(
                ApiCallRcImpl.simpleEntry(
                    ApiConsts.FAIL_UNKNOWN_ERROR,
                    "An IOException occurred while parsing SosFileReply"
                ),
                exc
            );
        }

        String nodeName = protoFilesReply.getNodeName();
        Path sosDir = tmpDirRef.resolve("tmp/" + nodeName);
        String sosDirStr = sosDir.toString();
        String sosReportDirOnStltStr = LinStor.SOS_REPORTS_DIR.resolve(sosReportNameRef).toString();
        for (File file : protoFilesReply.getFilesList())
        {
            String fileName = file.getFileName();
            if (fileName.startsWith(sosReportDirOnStltStr))
            {
                // move all '$stlt/var/lib/linstor.d/sos-report/$currentSosName/*' files to '$stlt/*'.
                fileName = fileName.substring(sosReportDirOnStltStr.length());
            }
            append(
                concatPaths(sosDirStr, fileName),
                file.getContent().toByteArray(),
                file.getTime()
            );
        }
        return requestNextBatch(tmpDirRef, sosReportNameRef, stltPeerRef, filesToRequestRef);
    }

    /**
     * After all files are processed, this method stores the controller-related information (if the
     * filtered node-list contains "Controller" ignoring case), tars the SOS report which also deletes the temporary
     * directory on the controller, sends out a cleanup message to all participating satellites (to delete their
     * temporary SOS-directory) and returns a Flux containing the sos report name.
     */
    private Flux<String> finishReport(
        Set<String> nodes,
        Set<String> rscs,
        Set<String> excludeNodes,
        Path tmpDir,
        String sosReportName,
        Date since,
        boolean includeCtrl,
        String queryParams
    )
        throws IOException, AccessDeniedException, ChildProcessTimeoutException, ExtCmdFailedException
    {
        String fileName = errorReporter.getLogDirectory() + "/" + sosReportName + SOS_SUFFIX;

        Stream<Node> nodeStream = getNodeStreamForSosReport(nodes, rscs, excludeNodes);
        List<Node> nodeList = nodeStream.collect(Collectors.toList());
        List<String> namesForTar = nodeList.stream().map(node -> node.getName().displayValue)
            .collect(Collectors.toList());
        for (String name : namesForTar)
        {
            try
            {
                makeDir(tmpDir.resolve(name));
            }
            catch (IOException exc)
            {
                throw new ApiRcException(
                    ApiCallRcImpl.simpleEntry(
                        ApiConsts.FAIL_UNKNOWN_ERROR,
                        "Failed to create directory: " + tmpDir.resolve(name)
                    ),
                    exc
                );
            }
        }
        if (includeCtrl)
        {
            namesForTar.add("_" + LinStor.CONTROLLER_MODULE);
            createControllerFilesInto(tmpDir, since);
        }
        namesForTar.add(QUERY_FILE);
        append(
            tmpDir.resolve("tmp/" + QUERY_FILE),
            (queryParams == null ? "<no query-params>" : queryParams).getBytes(),
            System.currentTimeMillis()
        );
        createTar(tmpDir, fileName, namesForTar);

        Flux<ApiCallRc> ret = Flux.empty();

        for (Node node : nodeList)
        {
            ret = ret.concatWith(
                node.getPeer(peerAccCtx.get()).apiCall(
                    InternalApiConsts.API_REQ_SOS_REPORT_CLEANUP,
                    stltComSerializer.headerlessBuilder()
                        .cleanupSosReport(sosReportName)
                        .build()
                ).map(
                    answer -> CtrlSatelliteUpdateCaller.deserializeApiCallRc(node.getName(), answer)
                )
                    // not sure why this is needed when we .thenMany later on... but without this, we get
                    // ClassCastExceptions as ApiCallRcImpl cannot be casted to String
                    .ignoreElements()
            );
        }

        return ret.thenMany(Flux.just(fileName));
    }

    /**
     * Creates a file with all supported and unsupported external tools of the given satellite in the given directory.
     */
    private void createExtToolsInfo(Path dirPath, Peer peer)
    {
        try
        {
            makeDir(dirPath);
            StringBuilder sb = new StringBuilder();
            ExtTools[] extTools = ExtTools.values();
            for (ExtTools tool : extTools)
            {
                ExtToolsInfo info = peer.getExtToolsManager().getExtToolInfo(tool);
                sb.append(tool.name()).append("\n");
                if (info.isSupported())
                {
                    sb.append("\tSupported\n");
                    sb.append("\tVersion: ")
                        .append(info.getVersionMajor())
                        .append(".")
                        .append(info.getVersionMinor())
                        .append(".")
                        .append(info.getVersionPatch())
                        .append("\n");
                }
                else
                {
                    sb.append("\tNot Supported\n");
                    sb.append("\tReason:\n");
                    for (String reason : info.getNotSupportedReasons())
                    {
                        sb.append("\t\t").append(reason).append("\n");
                    }
                }
            }
            Files.write(dirPath.resolve("extTools"), sb.toString().getBytes());
        }
        catch (IOException exc)
        {
            errorReporter.reportError(exc);
        }
    }

    /**
     * Collects all controller-based information (see list below) into the given <code>$tmpDir/_Controller</code>.
     *
     * Collected files/info:
     * <table>
     * <style>table tr td { padding-right: 10px; }</style>
     * <tr><th>Filename</th><th>Content</th></tr>
     * <tr><td>linstor.toml</td><td>'cp -p $actual_linstor.toml $sosDir'</td></tr>
     * <tr><td>journalctl</td><td>'journal -u linstor-controller --since $since'</td></tr>
     * <tr><td>ip-a</td><td>'ip a'</td></tr>
     * <tr><td>log-syslog</td><td>'cp -p /var/log/syslog $sosDir/log-syslog'</td></tr>
     * <tr><td>log-kern.log</td><td>'cp -p /var/log/kern.log $sosDir/log-kern.log'</td></tr>
     * <tr><td>log-messages</td><td>'cp -p /var/log/messages $sosdir/log-messages'</td></tr>
     * <tr><td>release</td><td>'cat /etc/redhat-release /etc/lsb-release /etc/os-release'</td></tr>
     * </table>
     */
    private void createControllerFilesInto(Path tmpDir, Date since) throws IOException
    {
        long nowMillis = System.currentTimeMillis();
        Date now = new Date(nowMillis);

        String nodeName = "_" + LinStor.CONTROLLER_MODULE;
        Path sosDir = tmpDir.resolve("tmp/" + nodeName);
        String infoContent = LinStor.linstorInfo() + "\n\nuname -a:           " + LinStor.getUname("-a");
        append(sosDir.resolve("linstorInfo"), infoContent.getBytes(), nowMillis);

        String timeContent = "Local Time: " + sdf.format(now) + "\nUTC Time:   " + sdfUtc.format(now);
        append(sosDir.resolve("timeInfo"), timeContent.getBytes(), nowMillis);
        getDbDump(sosDir);

        String tomlPath = ctrlCfg.getConfigDir() + LinstorConfig.LINSTOR_CTRL_CONFIG;
        CommandHelper[] commands = new CommandHelper[]
        {
            new CommandHelper(
                sosDir.resolve(LinstorConfig.LINSTOR_CTRL_CONFIG),
                new String[]
                {
                    "cp", "-p", tomlPath, sosDir.toString()
                }
            ),
            new CommandHelper(
                sosDir.resolve("journalctl"),
                new String[]
                {
                    "journalctl", "-u", "linstor-controller", "--since", LinStor.JOURNALCTL_DF.format(since)
                }
            ),
            new CommandHelper(
                sosDir.resolve("ip-a"),
                new String[]
                {
                    "ip", "a"
                }
            ),
            new CommandHelper(
                sosDir.resolve("log-syslog"),
                new String[]
                {
                    "cp", "-p", "/var/log/syslog", sosDir.toString() + "/log-syslog"
                }
            ),
            new CommandHelper(
                sosDir.resolve("log-kern.log"),
                new String[]
                {
                    "cp", "-p", "/var/log/kern.log", sosDir.toString() + "/log-kern.log"
                }
            ),
            new CommandHelper(
                sosDir.resolve("log-messages"),
                new String[]
                {
                    "cp", "-p", "/var/log/messages", sosDir.toString() + "/log-messages"
                }
            ),
            new CommandHelper(
                sosDir.resolve("release"),
                new String[]
                {
                    "cat", "/etc/redhat-release", "/etc/lsb-release", "/etc/os-release"
                }
            ),
        };
        for (CommandHelper cmd : commands)
        {
            makeFileFromCmdNoFailed(cmd.file, nowMillis, cmd.cmd);
        }

        FileCollector collector = new FileCollector();
        Files.walkFileTree(errorReporter.getLogDirectory(), collector);
        Set<SosReportType> errorReports = collector.getFiles();
        if (!errorReports.isEmpty())
        {
            makeDir(sosDir.resolve("logs/"));
        }

        for (SosReportType err : errorReports)
        {
            Path erroReportPath = sosDir.resolve("logs/" + Paths.get(err.getFileName()).getFileName());
            makeFileFromCmdNoFailed(
                addExtension(erroReportPath, ".out"),
                nowMillis,
                "cp", "-p",
                errorReporter.getLogDirectory().resolve(err.getFileName()).toString(),
                erroReportPath.toString()
            );
        }
    }

    private void makeDir(Path dirPath) throws IOException
    {
        if (!Files.exists(dirPath))
        {
            Files.createDirectory(dirPath);
        }
    }

    /**
     * Runs the given command and stores the stdOut in the given file.
     * The given file's lastModified date is set in the end.
     *
     * If a problem occurs, a new file with the additional ".err" is created with the content of the stdErr.
     * Another such special file is created for ".ioexc" (containing the stacktrace).
     *
     * If there is neither stdOut nor stdErr (and no exception) no files will be created at all (in case of a 'cp ...'
     * command for example)
     */
    private void makeFileFromCmdNoFailed(Path filePath, long timestamp, String... command)
    {
        LinkedBlockingDeque<Event> deque = new LinkedBlockingDeque<>(DFLT_DEQUE_CAPACITY);
        DaemonHandler dh = new DaemonHandler(deque, command);

        Path outFile = filePath;
        Path errFile = addExtension(filePath, ".err");

        boolean running;
        try
        {
            dh.startUndelimited();
            running = true;
        }
        catch (IOException exc)
        {
            running = false;

            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            exc.printStackTrace(new PrintWriter(baos));
            byte[] data = baos.toByteArray();

            append(addExtension(filePath, ".ioexc"), data, timestamp, command);
        }

        while (running)
        {
            Event event;

            byte[] data = null;
            Path file = null;
            try
            {
                event = deque.take();

                if (event instanceof StdOutEvent)
                {
                    StdOutEvent stdOutEvent = (StdOutEvent) event;
                    data = stdOutEvent.data;
                    file = outFile;
                }
                else if (event instanceof StdErrEvent)
                {
                    StdErrEvent stdErrEvent = (StdErrEvent) event;
                    data = stdErrEvent.data;
                    file = errFile;
                }
                else if (event instanceof EOFEvent)
                {
                    running = false;
                }
                else if (event instanceof ExceptionEvent)
                {
                    running = false;

                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    ((ExceptionEvent) event).exc.printStackTrace(new PrintWriter(baos));
                    data = baos.toByteArray();

                    file = addExtension(filePath, ".exc");
                }
                else
                {
                    errorReporter.logError(
                        "Unknown event type during SOS report: %s while processing command %s",
                        event.getClass().getCanonicalName(),
                        StringUtils.join(" ", command)
                    );
                }
            }
            catch (InterruptedException exc)
            {
                running = false;
                data = "Interrupted".getBytes();
                file = outFile;

                Thread.currentThread().interrupt();
            }
            if (file != null && data != null && data.length > 0)
            {
                append(file, data, timestamp, command);
            }
        }
    }

    private Path addExtension(Path pathRef, String extensionRef)
    {
        return pathRef.getParent().resolve(pathRef.getFileName() + extensionRef);
    }

    /**
     * Concatenates two paths with each other.
     * This is intentionally NOT done by using first.resolve(second) since if second is an absolute path the result of
     * the .resolve call is just plain the parameter (in this case $second).
     */
    private Path concatPaths(String first, String second)
    {
        return Paths.get(first, second);
    }

    /**
     * Appends the given data to the given path and setting the lastModifiedTime.
     * The given command is NOT executed, just needed for possible error reports.
     */
    private void append(Path file, byte[] data, long timestampRef, String... command)
    {
        try
        {
            if (!Files.exists(file.getParent()))
            {
                Files.createDirectories(file.getParent());
            }
            Files.write(file, data, StandardOpenOption.APPEND, StandardOpenOption.CREATE);
            errorReporter.logTrace("Written %d bytes to file %s", data.length, file.toString());
            Files.setAttribute(
                file,
                "lastModifiedTime",
                FileTime.fromMillis(timestampRef)
            );
        }
        catch (IOException exc)
        {
            if (command != null && command.length > 0)
            {
                errorReporter.logError(
                    "IOException occured while writing to %s from the command '%s'",
                    file.toString(),
                    StringUtils.join(" ", command)
                );
            }
            else
            {
                errorReporter.logError(
                    "IOException occured while writing to %s",
                    file.toString()
                );
            }
            errorReporter.reportError(exc);
        }

    }

    /**
     * Writes a database dump to $dirPath/dbDump
     */
    private void getDbDump(Path dirPath) throws IOException
    {
        try
        {
            Path dbDump = dirPath.resolve("dbDump");
            Files.write(dbDump, db.getDbDump().getBytes());
        }
        catch (DatabaseException exc)
        {
            String reportName = errorReporter.reportError(exc);
            append(
                dirPath.resolve("dbDump.failed"),
                ("ErrorReport-" + reportName).getBytes(),
                System.currentTimeMillis()
            );
        }
    }

    /**
     * Creates a $fileName.tar.gz file from the given source directory using the given directories relative to
     * '$source/tmp'.
     * The $source path is 'rm -rf'ed in the end.
     */
    private void createTar(Path source, String fileName, List<String> directories)
            throws IOException, ExtCmdFailedException, ChildProcessTimeoutException
    {
        ExtCmd extCommand = extCmdFactory.create();
        List<String> cmd = new ArrayList<>();
        cmd.add("tar");
        cmd.add("-C");
        cmd.add(source.toString() + "/tmp");
        cmd.add("-czvf");
        cmd.add(fileName);
        cmd.addAll(directories);
        String[] command = cmd.toArray(new String[0]);

        OutputData output = extCommand.exec(command);
        if (output.exitCode != 0)
        {
            throw new ExtCmdFailedException(command, output);
        }

        FileUtils.deleteDirectoryWithContent(source, errorReporter);
    }

    private Stream<Node> getNodeStreamForSosReport(Set<String> nodes, Set<String> rscs, Set<String> excludeNodes)
        throws AccessDeniedException
    {
        Set<Node> rscNodes = new HashSet<>();
        Set<String> copyRscs = rscs.stream().map(String::toLowerCase).collect(Collectors.toSet());
        if (!copyRscs.isEmpty())
        {
            for (ResourceDefinition rscDfn : rscDfnRepo.getMapForView(peerAccCtx.get()).values())
            {
                if (copyRscs.contains(rscDfn.getName().getDisplayName().toLowerCase()))
                {
                    rscDfn.streamResource(peerAccCtx.get()).forEach(rsc -> rscNodes.add(rsc.getNode()));
                }
            }
        }

        Set<String> copyNodes = nodes.stream().map(String::toLowerCase).collect(Collectors.toSet());
        Stream<Node> nodeStream = nodeRepository.getMapForView(peerAccCtx.get()).values().stream();
        if (!copyNodes.isEmpty() || !rscNodes.isEmpty())
        {
            nodeStream = nodeStream.filter(
                node -> copyNodes.contains(node.getName().getDisplayName().toLowerCase()) || rscNodes.contains(node)
            );
        }
        Set<String> copyExclude = excludeNodes.stream().map(String::toLowerCase).collect(Collectors.toSet());
        if (!copyExclude.isEmpty())
        {
            nodeStream = nodeStream
                .filter(node -> !copyExclude.contains(node.getName().getDisplayName().toLowerCase()));
        }
        return nodeStream;
    }

    private static class CommandHelper
    {
        private final Path file;
        private final String[] cmd;

        private CommandHelper(
            Path fileRef,
            String[] cmdRef
        )
        {
            file = fileRef;
            cmd = cmdRef;
        }
    }
}
